DOCKER_USERNAME = "jabriel"
PROJECT_NAME = "sneks"
SUFFIX = "-full-xarch"
DOCKER_IMAGE_PATTERN = (
    f"{DOCKER_USERNAME}/{PROJECT_NAME}:{{major}}.{{minor}}.{{micro}}{SUFFIX}"
)
SENV_NAME_PATTERN = f"{PROJECT_NAME}-{{major}}-{{minor}}-{{micro}}{SUFFIX}"

REQUIRED_PACKAGES = frozenset(
    ["dask", "distributed", "bokeh", "cloudpickle", "msgpack"]
)
# Most transitive dependencies of the required packages, so we don't have to re-install
# them with different versions according to the lockfile.
# Also include things we may want on the scheduler (`dask-pyspy`).
OPTIONAL_PACKAGES = frozenset(
    [
        "MarkupSafe",
        "click",
        "dask-pyspy",
        "fsspec",
        "heapdict",
        "jinja2",
        "locket",
        "lz4",
        "numpy",
        "packaging",
        "pandas",
        "partd",
        "psutil",
        "pyarrow",
        "pyparsing",
        "pyyaml",
        "s3fs",
        "sortedcontainers",
        "tblib",
        "toolz",
        "tornado",
        "urllib3",
        "zict",
    ]
)

if __name__ == "__main__":
    # Print constants as environment variables.
    # We can then read them into GitHub Actions with `python sneks/constants.py >> "$GITHUB_ENV"`

    for k, v in list(locals().items()):
        if k.isupper() and isinstance(v, str):
            print(f"{k}={v}")

    # Used in GitHub actions to give the correct image tag for a Python version,
    # given as CLI argument (the Python version in GHA might not be the version
    # we're building for)

    import sys

    if len(sys.argv) > 1:
        major, minor, micro = sys.argv[1].split(".")

        image = DOCKER_IMAGE_PATTERN.format(major=major, minor=minor, micro=micro)
        print(f"DOCKER_IMAGE={image}")



from __future__ import annotations

from typing import AbstractSet, Any


def current_versions_pdm(
    required: AbstractSet[str],
    optional: AbstractSet[str],
    lockfile: dict[str, Any],
    pyproject: dict[str, Any],
) -> tuple[list[str], list[str]]:
    """
    Determine what versions of required and optional packages are installed by parsing the lockfile

    Returns arguments to ``pip install`` to install those packages.

    Also validates that all dependencies in the lockfile will be installable on the cluster (no path deps, for example).
    """
    overrides: dict[str, str]
    try:
        overrides = pyproject["tool"]["pdm"]["overrides"]
    except KeyError:
        overrides = {}

    missing = set(required)
    pip_args: list[str] = []
    pip_overrides: list[str] = []
    for pkg in lockfile["package"]:
        name = pkg["name"]

        try:
            missing.remove(name)
        except KeyError:
            is_required = False
        else:
            is_required = True

        is_optional = (not is_required) and name in optional

        # TODO unlike Poetry, PDM doesn't annotate the lockfile with dependencies' categories.
        # So we can't verify that required packages aren't dev or optional deps.

        # Validate that we'll be able to install it, and ge the pip install arg if necessary.

        # But only for required packages, so as to not be too strict.
        if not (is_required or is_optional):
            continue

        if "path" in pkg and is_required:
            raise NotImplementedError(
                f"Required package {name!r} is installed as a path dependency. "
                "Uploading of local code is not supported yet. Please install from "
                "a Git URL, or use a version on PyPi."
            )

        if git_url := pkg.get("git"):
            # TODO what do other sources besides Git look like?
            pip_arg = f"git+{git_url}@{pkg['revision']}"
        else:
            pip_arg = f"{name}=={pkg['version']}"

        (pip_overrides if name in overrides else pip_args).append(pip_arg)

    if missing:
        missing = sorted(missing)
        raise ValueError(
            f"Required packages {missing} are not dependencies of your environment. "
            "These packages must be installed to create a dask cluster. "
            f"Run `pdm add {' '.join(missing)}` to install them."
        )

    assert not any(
        " " in o for o in overrides
    ), f"Overrides will fail; `sh` does not support arrays containing spaces: {overrides}"
    return pip_args, pip_overrides



from typing import Any, Callable, TypeVar, cast

from typing_extensions import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


def wraps_args(
    wrapped: Callable[P, Any]
) -> Callable[[Callable[..., T]], Callable[P, T]]:
    def inner(wrapper: Callable[..., T]) -> Callable[P, T]:
        import inspect

        wrapper.__dict__["__signature__"] = inspect.signature(wrapped).replace(
            return_annotation=inspect.signature(wrapper).return_annotation
        )
        return cast(Callable[P, T], wrapper)

    return inner



from __future__ import annotations

import subprocess
import sys

import coiled
import rich
from coiled.utils import parse_wait_for_workers
from distributed.client import Client
from distributed.nanny import Nanny
from distributed.worker import get_client as get_default_client

from sneks.compat import get_backend
from sneks.constants import DOCKER_IMAGE_PATTERN, SENV_NAME_PATTERN
from sneks.wraps_args import wraps_args


def _senv(account: str | None) -> str:
    """
    Make a Coiled container-only senv wrapping the Docker image.

    With the new senvs backend, cross-account senvs are no longer allowed, even
    container-only. Sadly, this means every sneks user has to create a separate sneks
    senv in their own account, pointing to the same image.

    Just re-creating the image every time is probably faster than checking if it exists.

    When ``Cluster`` gets an ``image=`` kwarg, we can do away with this silliness.
    """
    vi = sys.version_info
    senv = SENV_NAME_PATTERN.format(major=vi.major, minor=vi.minor, micro=vi.micro)
    coiled.create_software_environment(
        name=senv,
        container=DOCKER_IMAGE_PATTERN.format(
            major=vi.major, minor=vi.minor, micro=vi.micro
        ),
        account=account,
    )
    return senv


@wraps_args(coiled.Cluster)
def get_client(**kwargs) -> Client:
    """
    Launch a dask cluster in the cloud compatible with your current Poetry or PDM environment.

    All keyword arguments are forwarded to `coiled.Cluster`.

    You must be in the root directory of a Poetry or PDM project, with a ``pyproject.toml``
    and ``poetry.lock`` or ``pdm.lock`` file. All non-dev, non-optional dependencies listed in
    the lockfile will be installed on the cluster when you connect to it, then the workers
    will restart if necessary.
    """
    # TODO deal with async

    wait_for_workers = kwargs.pop("wait_for_workers", None)
    environ: dict[str, str] = kwargs.pop("environ", {})

    plugin, new_env = get_backend()
    environ.update(new_env)

    cluster = coiled.Cluster(
        software=_senv(kwargs.get("account")),
        environ=environ,
        **kwargs,
        wait_for_workers=False,
    )
    client = Client(cluster)
    rich.print(
        "[bold white]Uploading lockfile and installing dependencies on running workers[/]"
    )  # TODO improve
    try:
        client.register_worker_plugin(plugin)
    except subprocess.CalledProcessError as e:
        rich.print("[bold red]Dependency installation failed[/]")  # TODO improve
        rich.print("[stdout]", e.stdout.decode())
        rich.print("[stderr]", e.stderr.decode())
        raise

    if n_workers := kwargs.get("n_workers"):
        # Scale to requested size, if one was given. This is different from coiled behavior,
        # but ensures a cluster will look the way you're asking for it---more declarative style.
        rich.print(f"[bold white]Scaled to {n_workers} worker(s)[/]")  # TODO improve
        cluster.scale(n_workers)
    else:
        n_workers = cluster._start_n_workers

    target = parse_wait_for_workers(n_workers, wait_for_workers)
    rich.print(f"[bold white]Waiting for {target} worker(s)[/]")  # TODO improve
    client.wait_for_workers(target)

    # Hack around https://github.com/dask/distributed/issues/7035. Workers that joined while
    # the plugin was getting registered may have missed it.
    # We have to run this after we're sure all workers are connected.
    # If we just do it after the plugin is registered, it's not guaranteed that a _Worker_
    # subprocess that was starting up prior to `register_worker_plugin` has actually
    # started and connected to the scheduler yet.

    # don't want to serialize the whole plugin into the closure
    plugin_name = plugin.name

    async def restart_if_no_dep_manager(dask_worker: Nanny) -> None:
        if plugin_name not in dask_worker.plugins:
            print(f"{plugin_name!r} missing; fetching from scheduler...")

            msg = await dask_worker.scheduler.register_nanny()

            plugins = msg["nanny-plugins"]  # type: ignore
            if plugin_name not in plugins:
                raise RuntimeError(f"{plugin_name!r} not in {plugins}!")

            for name, plugin in plugins.items():
                print(f"Manually adding missing plugin {name!r}")
                await dask_worker.plugin_add(plugin=plugin, name=name)
        else:
            print(f"{plugin_name!r} already registered")

    # Sadly we have to block on this for consistency's sake.
    # We don't want to return control to the user until we're sure there
    # aren't any workers without the plugin.
    client.run(restart_if_no_dep_manager, nanny=True)

    # HACK: make the client "own" the cluster. When the client closes, the cluster
    # object will close too. Whether the actual Coiled cluster shuts down depends on the
    # `shutdown_on_close` argument.
    client._start_arg = None
    return client


# Utils for testing. Remove.
def can_import(module: str):
    import importlib

    importlib.import_module(module)
    return True


def check_import(module: str):
    get_default_client().submit(can_import, module, pure=False).result()



from __future__ import annotations

from typing import AbstractSet, Any


def current_versions_poetry(
    required: AbstractSet[str],
    optional: AbstractSet[str],
    lockfile: dict[str, Any],
    pyproject: dict[str, Any],
) -> tuple[list[str], list[str]]:
    """
    Determine what versions of required and optional packages are installed by parsing the lockfile

    Returns arguments to ``pip install`` to install those packages.

    Also validates that all dependencies in the lockfile will be installable on the cluster (no path deps, for example).
    """
    missing = set(required)
    pip_args: list[str] = []
    for pkg in lockfile["package"]:
        name = pkg["name"]

        try:
            missing.remove(name)
        except KeyError:
            is_required = False
        else:
            is_required = True

        is_optional = (not is_required) and name in optional

        if pkg.get("category", "") == "dev":
            if is_required:
                raise ValueError(
                    f"Required package {name} is only a dev dependency. "
                    "Dev dependencies will not be installed on the cluster.\n"
                    f"Upgrade this to a full dependency with `poetry add {name}`, or better, manually move it "
                    "from the `tool.poetry.dev-dependencies` section to the `tool.poetry.dependencies` section "
                    "of your `pyproject.toml`)."
                )
            continue

        if pkg.get("optional", False):
            if is_required:
                raise NotImplementedError(
                    f"Required package {name} is only an optional dependency. "
                    "Optional dependencies will not be installed on the cluster. "
                    "Please open an issue, since perhaps they should be.\n"
                    "For now, please make this a non-optional dependency."
                )
            continue

        # Reached here - package will be installed on cluster.
        # Validate that we'll be able to install it, and get the pip install arg if necessary.

        if source := pkg.get("source"):
            if (stype := source["type"]) != "git":
                if stype == "directory":
                    raise NotImplementedError(
                        f"Package {name!r} is installed as a path dependency. "
                        "Uploading of local code is not supported yet. Please install from "
                        "a Git URL, or use a version on PyPi."
                    )
                raise NotImplementedError(
                    f"Unsupported package source type {source['type']!r} for {name}. "
                    f"Please open an issue.\n{source=}"
                )
            if "subdirectory" in source:
                raise NotImplementedError(
                    "Subdirectory syntax of git repos not supported yet; "
                    "please open an issue."
                )
            # From git
            if is_required or is_optional:
                pip_args.append(f"git+{source['url']}@{source['resolved_reference']}")
        else:
            # From PyPi
            if is_required or is_optional:
                pip_args.append(f"{name}=={pkg['version']}")

    if missing:
        missing = sorted(missing)
        raise ValueError(
            f"Required packages {missing} are not dependencies of your environment. "
            "These packages must be installed to create a dask cluster. "
            f"Run `poetry add {' '.join(missing)}` to install them."
        )

    return pip_args, []



from __future__ import annotations

import asyncio
import gzip
import importlib
import logging
import os
import shutil
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING, ClassVar

import cloudpickle
from distributed.compatibility import to_thread  # type: ignore
from distributed.diagnostics.plugin import NannyPlugin

if TYPE_CHECKING:
    from distributed.nanny import Nanny

logger = logging.Logger(__name__)


class DepManagerBase(NannyPlugin, ABC):
    name: str = "DepManager"
    # ^ There should only ever be one instance of this plugin on a cluster at once.
    # So we always use the same name.
    _compressed_lockfile: bytes
    restart: ClassVar[bool] = False

    LOCKFILE_NAME: ClassVar[str]
    TOOL_NAME: ClassVar[str]

    def __init__(self, pyproject: bytes, lockfile: bytes) -> None:
        self._compressed_pyproject = gzip.compress(pyproject)
        self._compressed_lockfile = gzip.compress(lockfile)

    def get_tool_path(self) -> Path:
        "Get path to the installation tool"
        tool_path = Path.home() / ".local" / "bin" / self.TOOL_NAME.lower()
        if tool_path.is_file():
            return tool_path

        if tool_path_which := shutil.which(self.TOOL_NAME.lower()):
            return Path(tool_path_which)

        raise RuntimeError(f"cannot find {self.TOOL_NAME} installation")

    async def setup_tool(self, *, tool_path: Path, workdir: Path) -> None:
        "Do any setup needed before installation"
        return

    @abstractmethod
    async def install(self, *, tool_path: Path, workdir: Path) -> bool:
        "Do installation, return whether to restart"
        raise NotImplementedError

    async def setup(self, nanny: Nanny) -> None:
        workdir = Path(nanny.local_directory)
        pyproject_path = workdir / "pyproject.toml"
        lockfile_path = workdir / self.LOCKFILE_NAME
        await asyncio.gather(
            write_compressed_file(self._compressed_pyproject, pyproject_path),
            write_compressed_file(self._compressed_lockfile, lockfile_path),
        )

        # TODO skip installation and don't restart if there's already a lockfile and it's up to date.

        tool_path = self.get_tool_path().absolute()
        print(f"{self.TOOL_NAME} available at {tool_path}")

        await self.setup_tool(tool_path=tool_path, workdir=workdir)

        restart = await self.install(tool_path=tool_path, workdir=workdir)
        if not restart:
            print("No dependencies to install or update")
            return

        print("Installation complete! Restarting worker.")
        # NOTE: setting `self.restart` _should_ work, but there are probably race conditions between
        # `Client.restart` and nannies resetarting themselves. The `Nanny.restart` method is generally
        # not very trustworthy.
        await nanny.kill(timeout=30)

    def __getstate__(self) -> dict:
        """
        Make this object unpickleable to dask without its package being installed on workers, via a horrible hack.

        This avoids needing `sneks` to actually be installed on the cluster, just on the client side.
        """
        # HACK, awful horrible hack.
        # The implementation of `distributed.protocol.pickle.dumps` currently tries to plain `pickle.dumps`
        # the object. If that raises an error, then it tries `cloudpickle.dumps`.
        # We game this code structure by raising an error every other time we're pickled, forcing the
        # `cloudpickle.dumps` codepath and then succeeding on that one.
        if getattr(self, "_the_first_pickle_is_the_deepest", True):
            self._the_first_pickle_is_the_deepest = False
            raise RuntimeError("cloudpickle time!")
        else:
            del self._the_first_pickle_is_the_deepest

            # Very awkward, but the only way to tell cloudpickle to pickle something by value
            # is by passing it the module whose members you want pickled by value.
            module = importlib.import_module(self.__module__)
            cloudpickle.register_pickle_by_value(module)

            return self.__dict__

    @staticmethod
    async def run(
        program: str | Path,
        *args: str,
        tee: bool = True,
        cwd: Path | None = None,
        env: dict[str, str] | None = None,
    ) -> tuple[bytes, bytes]:
        call = f"{program} {' '.join(args)}"
        print(f"Executing {call}")
        proc = await asyncio.create_subprocess_exec(
            program,
            *args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=cwd,
            env=env,
        )

        stdout, stderr = await proc.communicate()
        returncode = proc.returncode
        assert returncode is not None

        print(f"{call} exited with {returncode}")
        if tee:
            if stdout:
                sys.stdout.write(stdout.decode())
            if stderr:
                sys.stderr.write(stderr.decode())

        if returncode != 0:
            raise CalledProcessError(returncode, call, stdout, stderr)
        return stdout, stderr


async def write_compressed_file(data: bytes, path: Path) -> None:
    def _write():
        with open(path, "wb") as f:
            f.write(gzip.decompress(data))

    await to_thread(_write)
    print(f"Wrote to {path}")


# Concrete implementations
##########################
#
# Sadly, these all have to be in the same file for pickle-by-value to work.
# Otherwise, they'd still reference an import from sneks for the base class.


class PoetryDepManager(DepManagerBase):
    LOCKFILE_NAME: ClassVar[str] = "poetry.lock"
    TOOL_NAME: ClassVar[str] = "Poetry"

    async def setup_tool(self, *, tool_path: Path, workdir: Path) -> None:
        "Make poetry use the global environment"
        # TODO do this without a subprocess
        # TODO only do this once
        await self.run(tool_path, "config", "virtualenvs.create", "false")
        print("Poetry configured to use global environment")

    async def install(self, *, tool_path: Path, workdir: Path) -> bool:
        cwd = Path.cwd()
        try:
            os.chdir(workdir)
            out, err = await self.run(
                tool_path,
                "install",
                "--sync",
                "--only=main",
                "--no-root",
                "-n",
            )
            # HACK we can do better than this text parsing to decide
            # whether to restart or not
            return b"Updating" in out or b"Removing" in out
        finally:
            os.chdir(cwd)


class PdmDepManager(DepManagerBase):
    LOCKFILE_NAME: ClassVar[str] = "pdm.lock"
    TOOL_NAME: ClassVar[str] = "PDM"

    async def install(self, *, tool_path: Path, workdir: Path) -> bool:
        await self.run(
            tool_path,
            "info",
            cwd=workdir,
        )
        out, err = await self.run(
            tool_path,
            "sync",
            "--prod",
            "--clean",
            "--no-self",
            "--no-isolation",
            cwd=workdir,
        )
        # HACK we can do better than this text parsing to decide
        # whether to restart or not
        return b"nothing to do" not in out



from __future__ import annotations

from pathlib import Path
from typing import AbstractSet, Any, Callable

import tomli

from sneks.constants import OPTIONAL_PACKAGES, REQUIRED_PACKAGES
from sneks.parse_pdm import current_versions_pdm
from sneks.parse_poetry import current_versions_poetry
from sneks.plugin import DepManagerBase, PdmDepManager, PoetryDepManager


def find_pyproject() -> Path:
    # TODO actually implement this
    return Path("pyproject.toml")


def sniff_tool_type(pyproject: dict[str, Any]) -> str:
    # TODO handle no build system
    backend: str = pyproject["build-system"]["build-backend"]
    return backend.split(".", maxsplit=1)[0]


def get_backend() -> tuple[DepManagerBase, dict[str, str]]:
    "Get the `DepManagerBase` plugin instance, and extra environment variables"
    pyproject_path = find_pyproject()
    with open(pyproject_path, "rb") as f:
        pyproject = f.read()

    # TODO handle bad TOML
    pyproject_data = tomli.loads(pyproject.decode())
    tool = sniff_tool_type(pyproject_data)

    plugin_type: type[DepManagerBase]
    current_versions_from_lockfile: Callable[
        [AbstractSet[str], AbstractSet[str], dict[str, Any], dict[str, Any]],
        tuple[list[str], list[str]],
    ]
    if tool == "poetry":
        plugin_type = PoetryDepManager
        current_versions_from_lockfile = current_versions_poetry
    elif tool == "pdm":
        plugin_type = PdmDepManager
        current_versions_from_lockfile = current_versions_pdm
    else:
        raise ValueError(f"Unsupported build tool {tool}")

    lockfile_path = pyproject_path.parent / plugin_type.LOCKFILE_NAME
    with open(lockfile_path, "rb") as f:
        lockfile = f.read()
    # TODO handle bad TOML
    lockfile_data = tomli.loads(lockfile.decode())

    required_versions, overrides = current_versions_from_lockfile(
        REQUIRED_PACKAGES, OPTIONAL_PACKAGES, lockfile_data, pyproject_data
    )
    environ = {
        "PIP_PACKAGES": " ".join(required_versions),
        "PIP_OVERRIDES": " ".join(overrides),
    }
    return plugin_type(pyproject, lockfile), environ
